//#undef USE_2P
#define USE_2P
//#undef USE_ISEND
//#undef USE_FLOWCTL
#define USE_FLOWCTL

#define SEND_TIME
#define PERFORMANCE
//#undef PERF_SHORT
//#undef DELAY
//#undef FREQUENCIES

//#undef CORRECTNESS

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)
//#define gamma_ISEND(a,b,c)  gamma_isend_flowctl(a,b,c)
//#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p_flowctl(a,b,c,d,e)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p(a,b,c,d,e)
//#define gamma_ISEND(a,b,c)  gamma_isend(a,b,c)
//#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p(a,b,c,d,e)
#endif

#ifdef USE_2P
#define LEN_1P 12
#endif

#include <linux/gamma/libgamma.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

#undef TRACE
//#define TRACE
#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 5


unsigned char *data,*data1;
int len;

int main(int argc, char **argv) {

#ifdef PERFORMANCE
#ifdef SEND_TIME
  double record_mc[NPROVE];
#endif
  double record_mm[NPROVE];
#endif
  int nprocs,i,me;
  time_586
    t_send_s,
    t_rec;
#ifdef SEND_TIME
  time_586
    t_send_e;
#endif
#ifdef PERFORMANCE
  double time_overhead;
#endif

  nprocs = atoi(argv[2]);
  gamma_init(nprocs, argc, argv);
  len = atoi(argv[1]);

  me = gamma_my_node();
  /*  Allocate data space (+4096 to allow page alignment)  */
#ifdef USE_2P
  len = len+LEN_1P;
#endif
  data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
  /*  Align to page boundary  */
  data = (unsigned char *)(((unsigned long)data1+4096) & (~4095));


  if (me == MASTER_NODE)
    {
      for (i=0; i<nprocs-1; i++)
	gamma_set_passive_port(
			       PORT+i+1,
			       BROADCAST, gamma_my_par_pid(), PORT,
			       PORT, GO_BACK, data, len
			       );
      gamma_set_passive_port(
			     PORT,
			     BROADCAST, gamma_my_par_pid(), PORT,
			     PORT, GO_BACK, data, len
			     );
    }
  else
    gamma_set_passive_port(
			   PORT,
			   MASTER_NODE, gamma_my_par_pid(), PORT+me,
			   PORT, GO_BACK, data, len
			   );
  

#ifdef PERFORMANCE
  /*  Computing the gamma_time() overhead...  */
  for (i=0; i<NPROVE; i++) 
    {
      gamma_time(t_send_s);
      gamma_time(t_rec);
      record_mm[i] = gamma_time_diff(t_rec, t_send_s);
    }
  {
    double Mmm,medmm,varmm;
    mean_variance(record_mm, &time_overhead ,&Mmm, &medmm, &varmm);
  }
#endif

  //gamma_sync();

  for (i=0; i<NPROVE; i++) 
    {
      if (me == MASTER_NODE) 
	{
	  //gamma_sync();
	  //sleep(1);

#ifdef PERFORMANCE
	  gamma_time(t_send_s);
#endif

#ifdef USE_2P
	  gammaSEND2P(PORT, data, LEN_1P, data+LEN_1P, len-LEN_1P);
#else
	  gamma_SEND(PORT, data, len);
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
	  gamma_time(t_send_e);
#endif
#endif
#ifdef TRACE
	  printf("[%d][%d]: master sent\n", me, i);
	  fflush(stdout);
#endif
	  gamma_wait(PORT, nprocs-1);

#ifdef TRACE
	  printf("[%d][%d]: master recvd\n", me, i);
	  fflush(stdout);
#endif
#ifdef PERFORMANCE
	  gamma_time(t_rec);
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
	  record_mc[i] =
	    gamma_time_diff(t_send_e, t_send_s) - time_overhead;
#endif
	  record_mm[i] =
	    gamma_time_diff(t_rec, t_send_s) - time_overhead;
#endif
	}
      else 
	{
	  /*
	  printf("[%d][%d]: slave waiting\n",me, i);
	  fflush(stdout);*/
	  gamma_wait(PORT, 1);
#ifdef TRACE
	  printf("[%d][%d]: slave recvd\n", me, i);
	  fflush(stdout);
#endif
#ifdef USE_2P
	  gammaSEND2P(PORT, data, LEN_1P, data+LEN_1P, len-LEN_1P);
#else
	  gamma_SEND(PORT, data, len);
#endif
#ifdef TRACE
	  printf("[%d][%d]: slave sent\n", me, i);
	  fflush(stdout);
#endif
	}
    }
  /*
   *  Force printout of kernel-level time measurements,
   *  if enabled.  Otherwise nothing happens.
   */
  gamma_timers(t_send_s,t_rec);

  free(data1);
  gamma_exit();

#ifdef PERFORMANCE
  if (me == MASTER_NODE) 
    {
#ifdef SEND_TIME
      double mmc,Mmc,medmc,varmc;
#endif
      double mmm,Mmm,medmm,varmm;
      printf("\n\n");
      printf("PERFORMANCE (microsec, MB/sec) (lowest < average < highest):\n\n");
#ifdef SEND_TIME
      mean_variance(record_mc, &mmc, &Mmc, &medmc, &varmc);
      printf("total send time        %g < %g < %g\n",
	     mmc, medmc, Mmc);
      if (len)
	printf("send throughput        %g < %g < %g\n",
	       len/Mmc, len/medmc,
	       len/mmc);
#endif
      mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
      printf("total delay            %g < %g < %g\n",
	     mmm, medmm, Mmm);
      if (len)
	printf("end-to-end throughput  %g < %g < %g\n",
	       len/Mmm, len/medmm,
	       len/mmm);
      printf("\n\n");

      printf("VARIANCE:  \n");
#ifdef SEND_TIME
      printf("send time        %g\n", varmc);
#endif
      printf("end-to-end delay %g\n", varmm);
      printf("\n\n");
	       
      fflush(stdout);    
    }
#endif
  return 0;
}
