/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver e1000_main.c

    Copyright		(C) 2006 Giuseppe Ciaccio
    e-mail:		ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/*
				 * Max waiting (usec) for each pending Tx ring
				 * entry to be marked as "done", while waiting
				 * for Tx ring to flush
				 */

#define GAMMA_TX_RING_SIZE 256	/*
				 *  Size of Tx descriptor ring.
				 *  PRO/1000: must be multiple of 8
				 */
#define GAMMA_RX_RING_SIZE 4096	/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 *  PRO/1000: must be multiple of 8
				 */

#define MAX_CREDIT_SIZE 32
#define LOW_WATERMARK 28	/*  Must be > 0  */
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */

/*  NOTE:  Jumbo frames larger than 2072 can trigger a NIC hardware bug.  */
#ifdef USE_JUMBO_FRAMES
#define FRAME_SIZ 2072		/*  Max payload chunk = 2 Kbytes  */
#else
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#endif


#ifdef MORE_NICS_SAMENODE_ROUTING
/*======================================================================*/
/*  GLOBAL MACROS                                                       */
/*======================================================================*/

/* these global macros are currently used only by rx macros */

#define bind2_global_decl \
struct eg1000_adapter *global_adapters[MAX_NUM_NICS]; \
struct eg1000_tx_ring *global_rx_rings[MAX_NUM_NICS]

#define bind2_global_init \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) {\
		global_adapters[i] = gamma_devices[i]->priv; \
		global_rx_rings[i] = global_adapters[i]->rx_ring; \
	} \
} while (0)

#endif


/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl \
	struct eg1000_adapter *adapter = NULL; \
	struct eg1000_tx_ring *tx_ring = NULL; \
	struct eg1000_tx_desc *tx_desc; \
	uint32_t txd_upper = 0, txd_lower = E1000_TXD_CMD_IFCS; \
	unsigned int idx

#define tx_common_init(nic_idx) \
do {\
	adapter = gamma_devices[nic_idx]->priv; \
	tx_ring = adapter->tx_ring; \
	idx = tx_ring->next_to_use; \
} while (0)

#define tx_ring_head (tx_ring->next_to_clean)
#define tx_ring_tail (tx_ring->next_to_use)
#define tx_ring_occupancy \
((tx_ring->next_to_clean <= tx_ring->next_to_use) ? \
tx_ring->next_to_use - tx_ring->next_to_clean : \
GAMMA_TX_RING_SIZE - (tx_ring->next_to_clean - tx_ring->next_to_use))

#define tx_reset_and_restart() \
do { eg1000_down(adapter); eg1000_up(adapter); busywait(7000000); } while (0)

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
do { \
idx = tx_ring->next_to_use; \
\
tx_desc = E1000_TX_DESC(*tx_ring, idx); \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
tx_desc->buffer_addr = cpu_to_le64(virt_to_phys(pkt_ptr)); \
tx_desc->lower.data = cpu_to_le32( \
  txd_lower | gmax((pkt_size),60)) | cpu_to_le32(adapter->txd_cmd); \
tx_desc->upper.data = cpu_to_le32(txd_upper); \
incr_modulo(idx,GAMMA_TX_RING_SIZE); \
wmb(); \
} while (0)

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
\
if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
  memcpy(&(((unsigned char *)header_ptr)[header_size]), \
         (void *)phys_to_virt((unsigned long)(payload_ptr)), \
         payload_size); \
  tx_enq_pkt1(header_ptr,(header_size)+(payload_size)); \
  break; \
} \
\
idx = tx_ring->next_to_use; \
tx_desc = E1000_TX_DESC(*tx_ring, idx); \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
tx_desc->buffer_addr = cpu_to_le64(virt_to_phys(header_ptr)); \
tx_desc->lower.data = cpu_to_le32(txd_lower | (header_size)); \
tx_desc->upper.data = cpu_to_le32(txd_upper); \
incr_modulo(idx,GAMMA_TX_RING_SIZE); \
\
tx_desc = E1000_TX_DESC(*tx_ring, idx); \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
tx_desc->buffer_addr = cpu_to_le64((unsigned long)payload_ptr);\
tx_desc->lower.data = cpu_to_le32( \
  txd_lower | gmax((payload_size),60)) | cpu_to_le32(adapter->txd_cmd); \
tx_desc->upper.data = cpu_to_le32(txd_upper); \
incr_modulo(idx,GAMMA_TX_RING_SIZE); \
\
wmb(); \
break; \
}

#define tx_start() E1000_WRITE_REG(&adapter->hw, TDT, idx)

#define tx_forw_tail() tx_ring->next_to_use = idx

#define tx_flush_ring(tx_desc_idx,n,res) \
do { \
time_586 a,b; \
int r; \
\
res = r = 1; \
for (;;) { \
  if (tx_ring_head == tx_ring_tail) /* empty ring */ \
    break; \
  tx_desc = E1000_TX_DESC(*tx_ring,tx_ring_head); \
  /* skip non-EOP Tx descs */ \
  if (!(tx_desc->lower.data & cpu_to_le32(E1000_TXD_CMD_EOP))) { \
    incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
    continue; \
  } \
  gamma_time(a); \
  for (;;) { \
    if (tx_desc->upper.data & cpu_to_le32(E1000_TXD_STAT_DD)) { \
      tx_desc->buffer_addr = 0; \
      tx_desc->lower.data = 0; \
      tx_desc->upper.data = 0; \
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      break; \
    } \
    gamma_time(b); \
    if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
      r = 0; /* tx desc timed out */ \
      break; \
    } \
  } \
  if (!r) { \
    printg( \
       "tx_flush_ring(): failure at pos %d, status %8.8x (tail is at %d).\n", \
        tx_ring_head,tx_desc->upper.data,tx_ring_tail); \
    tx_desc->buffer_addr = 0; \
    tx_desc->lower.data = 0; \
    tx_desc->upper.data = 0; \
    /* res = 0; */ /*  report the failure  */ \
    tx_reset_and_restart(); \
    printg("tx_flush_ring(): NIC restarted\n"); \
    /*  tx_ring_head and tx_ring_tail are now set to 0  */ \
    break; \
  } \
  if (tx_ring_head == (tx_desc_idx)) \
    break; \
}\
} while (0)

/*  similar to tx_flush_ring(), different guard; assumes non-empty ring  */
#define tx_wait_ring_room(n,res) \
do { \
time_586 a,b; \
int r; \
\
res = r = 1; \
for (;;) { \
  tx_desc = E1000_TX_DESC(*tx_ring,tx_ring_head); \
  /* skip non-EOP Tx descs */ \
  if (!(tx_desc->lower.data & cpu_to_le32(E1000_TXD_CMD_EOP))) { \
    incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
    continue; \
  } \
  gamma_time(a); \
  for (;;) { \
    if (tx_desc->upper.data & cpu_to_le32(E1000_TXD_STAT_DD)) { \
      tx_desc->buffer_addr = 0; \
      tx_desc->lower.data = 0; \
      tx_desc->upper.data = 0; \
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      break; \
    } \
    gamma_time(b); \
    if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
      r = 0; /* tx desc timed out */ \
      break; \
    } \
  } \
  if (!r) { \
    printg( \
      "tx_wait_ring_room(): failure at pos %d, status %8.8x (tail is at %d).\n", \
      tx_ring_head,tx_desc->upper.data,tx_ring_tail); \
    tx_desc->buffer_addr = 0; \
    tx_desc->lower.data = 0; \
    tx_desc->upper.data = 0; \
    /* res = 0; */ /*  report the failure  */ \
    tx_reset_and_restart(); \
    printg("tx_wait_ring_room(): NIC restarted\n"); \
    /*  tx_ring_head and tx_ring_tail are now set to 0  */ \
    break; \
  } \
/*  if (tx_ring_has_room()) */ if (tx_ring_head == tx_ring_tail) \
    break; \
}\
} while (0)


/*======================================================================*/
/*  RX MACROS                                             	        */
/*======================================================================*/

#define rx_common_decl_and_init \
	struct eg1000_adapter *adapter = dev->priv; \
	struct eg1000_rx_ring *rx_ring = adapter->rx_ring; \
	struct eg1000_rx_desc *rx_desc

#ifdef MORE_NICS_SAMENODE_ROUTING
#define poll_common_decl \
	struct eg1000_rx_desc *rx_desc
#else
#define poll_common_decl_and_init \
	struct eg1000_adapter *adapter = gamma_devices[0]->priv; \
	struct eg1000_rx_ring *rx_ring = adapter->rx_ring; \
	struct eg1000_rx_desc *rx_desc
#endif

#define rx_ring_tail 0
#define rx_ring_head rx_ring->next_to_clean
#define read_rx_ring_head(head) (head) = rx_ring->next_to_clean
#define write_rx_ring_head(head) rx_ring->next_to_clean = (head)
#define forward_rx_ring_head() \
do { \
  incr_modulo(rx_ring->next_to_clean,GAMMA_RX_RING_SIZE); \
  incr_modulo(rx_ring->next_to_use,GAMMA_RX_RING_SIZE); \
} while (0)

#ifdef MORE_NICS_SAMENODE_ROUTING

#define rx_ring_head_arg(idx) global_rx_rings[idx]->next_to_clean

#define rx_read_pkt_info_light_arg(idx) \
do { \
  rx_desc = E1000_RX_DESC(*global_rx_rings[idx],rx_ring_head_arg(idx)); \
} while (0)

#define rx_read_pkt_info_arg(idx) \
do { \
  rx_desc = E1000_RX_DESC(*global_rx_rings[idx],rx_ring_head_arg(idx)); \
} while (0)

#endif

#define clear_rx_irqs()

#define rx_read_pkt_info_light() \
do { \
  rx_desc = E1000_RX_DESC(*rx_ring,rx_ring_head); \
} while (0)

#define rx_read_pkt_info() \
do { \
  rx_desc = E1000_RX_DESC(*rx_ring,rx_ring_head); \
} while (0)

#define rx_status() rx_desc->errors
#define rx_err_pktnotavail() \
(!(rx_desc && rx_desc->status & E1000_RXD_STAT_DD))

#define discard_pkt() \
do { \
  rx_desc->status = 0; \
  wmb(); \
  E1000_WRITE_REG(&adapter->hw, RDT, rx_ring_head); \
  forward_rx_ring_head(); \
} while (0)

#define rx_header(h) \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
(h) = (l2s *)phys_to_virt(le64_to_cpu(rx_desc->buffer_addr))

#define invoke_standard_rx(dev) eg1000_clean_rx_irq(dev->priv,rx_ring)

#define rx_suspend_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		eg1000_irq_disable(gamma_devices[i]->priv); \
} while (0)

#define rx_resume_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		eg1000_irq_enable(gamma_devices[i]->priv); \
} while (0)
