/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver tg3.c

    Copyright		(C) 2006 Giuseppe Ciaccio
    e-mail:		ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000  /*
                                 * Max waiting (usec) for each pending Tx ring
				 * entry to be marked as "done", while waiting
				 * for Tx ring to flush
				 */
#define GAMMA_TX_RING_SIZE 512	/*
				 *  Size of Tx descriptor ring.
				 *  tg3: hardcoded, do NOT change!
				 */
#define GAMMA_RX_RING_SIZE TG3_RX_RCB_RING_SIZE(tp)
				/*
				 *  Size of Rx descriptor ring.
				 *  tg3: hardcoded, do NOT change!
				 */

#define MAX_CREDIT_SIZE 32
#define LOW_WATERMARK 28	/*  Must be > 0  */
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */

#ifdef USE_JUMBO_FRAMES
#define FRAME_SIZ 4120		/*  Max payload chunk = 4 Kbytes  */
#else
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#endif


#ifdef MORE_NICS_SAMENODE_ROUTING
/*======================================================================*/
/*  GLOBAL MACROS                                                       */
/*======================================================================*/

/* these global macros are currently used only by rx macros */

#define bind2_global_decl \
struct eg1000_adapter *global_adapters[MAX_NUM_NICS]; \
struct eg1000_desc_ring *global_rx_rings[MAX_NUM_NICS]

#define bind2_global_init \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) {\
		global_adapters[i] = gamma_devices[i]->priv; \
		global_rx_rings[i] = &global_adapters[i]->rx_ring; \
	} \
} while (0)

#endif


/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl \
	struct tg3 *tp; \
	struct tgg3_tx_buffer_desc *txd; \
	dma_addr_t addr; \
	u32 tg3_entry

#define tx_common_init(nic_idx) \
do { tp = netdev_priv(gamma_devices[nic_idx]); } while (0)

#define tx_ring_head (tp->tx_cons)
#define tx_ring_tail (tp->tx_prod)
#define tx_ring_occupancy \
((tp->tx_cons <= tp->tx_prod) ? \
tp->tx_prod - tp->tx_cons : GAMMA_TX_RING_SIZE - (tp->tx_cons - tp->tx_prod))

/* peppe da fare? */
#define tx_reset_and_restart()

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
do { \
  tg3_entry = tp->tx_prod; \
  txd = &tp->tx_ring[tg3_entry]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  addr = virt_to_phys(pkt_ptr); \
  txd->addr_hi = (u64)addr >> 32; \
  txd->addr_lo = (u64)addr & 0xffffffff; \
  txd->len_flags = ((gmax((pkt_size),60)) << TXD_LEN_SHIFT) | TXD_FLAG_END; \
  txd->vlan_tag = 0; \
  incr_modulo(tg3_entry,GAMMA_TX_RING_SIZE); \
  wmb(); \
} while (0)

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
\
  if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
    memcpy(&(((unsigned char *)header_ptr)[header_size]), \
            (void *)phys_to_virt((unsigned long)(payload_ptr)),payload_size); \
    tx_enq_pkt1(header_ptr,(header_size)+(payload_size)); \
    break; \
  } \
\
  tg3_entry = tp->tx_prod; \
  txd = &tp->tx_ring[tg3_entry]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  addr = virt_to_phys(header_ptr); \
  txd->addr_hi = (u64)addr >> 32; \
  txd->addr_lo = (u64)addr & 0xffffffff; \
  txd->len_flags = ((header_size) << TXD_LEN_SHIFT); \
  txd->vlan_tag = 0; \
  incr_modulo(tg3_entry,GAMMA_TX_RING_SIZE); \
\
  txd = &tp->tx_ring[tg3_entry]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  txd->addr_hi = (u64)(unsigned long)payload_ptr >> 32; \
  txd->addr_lo = (u64)(unsigned long)payload_ptr & 0xffffffff; \
  txd->len_flags = ((payload_size) << TXD_LEN_SHIFT) | TXD_FLAG_END; \
  txd->vlan_tag = 0; \
  incr_modulo(tg3_entry,GAMMA_TX_RING_SIZE); \
\
  wmb(); \
  break; \
}

#define tx_start() \
tw32_tx_mbox((MAILBOX_SNDHOST_PROD_IDX_0 + TG3_64BIT_REG_LOW), tg3_entry)

#define tx_forw_tail() tp->tx_prod = tg3_entry

#define tx_flush_ring(tx_desc_idx,n,res) \
do { \
  time_586 a,b; \
  int r; \
\
  res = r = 1; \
  for (;;) { \
    if (tx_ring_head == tx_ring_tail) /* empty ring */ break; \
    gamma_time(a); \
    for (;;) { \
      if (tx_ring_head != tp->hw_status->idx[0].tx_consumer) { \
	incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
        break; \
      } \
      gamma_time(b); \
      if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
        r = 0; /* tx desc timed out */ \
        break; \
      } \
    } \
    if (!r) { \
      printg( \
         "tx_flush_ring(): failure at pos %d, status %08x (tail at %d).\n", \
          tx_ring_head,tp->hw_status->status,tx_ring_tail); \
      res = 0; /*  report the failure  */ \
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      r = 1; \
    } \
    if (tx_ring_head == (tx_desc_idx)) break; \
  }\
} while (0)

/*  similar to tx_flush_ring(), different guard; assumes non-empty ring  */
#define tx_wait_ring_room(n,res) \
do { \
  time_586 a,b; \
  int r; \
\
  res = r = 1; \
  for (;;) { \
    gamma_time(a); \
    for (;;) { \
      if (tx_ring_head != tp->hw_status->idx[0].tx_consumer) { \
        incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
        break; \
      } \
      gamma_time(b); \
      if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
        r = 0; /* tx desc timed out */ \
        break; \
      } \
    } \
    if (!r) { \
      printg( \
         "tx_wait_ring_room(): failure at pos %d, status %08x (tail at %d).\n", \
          tx_ring_head,tp->hw_status->status,tx_ring_tail); \
      res = 0; /*  report the failure  */ \
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      r = 1; \
    } \
    /* if (tx_ring_has_room()) */ if (tx_ring_head == tx_ring_tail) break; \
  }\
} while (0)


/*======================================================================*/
/*  RX MACROS                                             	        */
/*======================================================================*/

#define rx_common_decl_and_init \
	struct tg3 *tp = netdev_priv(dev); \
	struct tgg3_rx_buffer_desc *desc = NULL,*rdesc; \
	u32 opaque_key = 0, desc_idx, *post_ptr = NULL; \
	u16 hw_idx

#ifdef MORE_NICS_SAMENODE_ROUTING
#define poll_common_decl \
	?????
#else
#define poll_common_decl_and_init \
	struct tg3 *tp = netdev_priv(gamma_devices[0]); \
	struct tgg3_rx_buffer_desc *desc,*rdesc; \
	u32 opaque_key, desc_idx, *post_ptr; \
	u16 hw_idx
#endif

#define rx_ring_tail hw_idx
#define rx_ring_head (tp->rx_rcb_ptr)
#define read_rx_ring_head(head) (head) = rx_ring_head
#define write_rx_ring_head(head)  rx_ring_head = (head)
#define forward_rx_ring_head() \
do { incr_modulo(rx_ring_head,GAMMA_RX_RING_SIZE); } while (0)

#ifdef MORE_NICS_SAMENODE_ROUTING

#define rx_ring_head_arg(idx) global_rx_rings[idx]->rx_rcb_ptr ???
#define rx_read_pkt_info_arg(idx) ????? \
do { \
} while (0)

#endif

#define clear_rx_irqs()
//tw32_mailbox(MAILBOX_INTERRUPT_0 + TG3_64BIT_REG_LOW, 0x00000000)

#define rx_read_pkt_info_light() \
do { \
  hw_idx = tp->hw_status->idx[0].rx_producer; \
  rmb(); \
} while (0)

#define rx_read_pkt_info() \
do { \
  hw_idx = tp->hw_status->idx[0].rx_producer; \
  rmb(); \
  for (;;) { \
    if (rx_ring_head == hw_idx) break; \
    rdesc = &tp->rx_rcb[rx_ring_head]; \
    desc_idx = rdesc->opaque & RXD_OPAQUE_INDEX_MASK; \
    opaque_key = rdesc->opaque & RXD_OPAQUE_RING_MASK; \
    if (opaque_key == RXD_OPAQUE_RING_STD) { \
      desc = &tp->rx_std[desc_idx]; \
      post_ptr = &tp->rx_std_ptr; \
      break; \
    } else if (opaque_key == RXD_OPAQUE_RING_JUMBO) { \
      desc = &tp->rx_jumbo[desc_idx]; \
      post_ptr = &tp->rx_jumbo_ptr; \
      break; \
    } else { \
      forward_rx_ring_head(); \
    } \
  } \
} while (0)

#define rx_status() (rdesc ? rdesc->err_vlan & RXD_ERR_MASK: 0)
#define rx_err_pktnotavail() (rx_ring_head == hw_idx)

#define rx_header(h) \
do { \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  (h) = (l2s *)phys_to_virt(((u64)desc->addr_hi << 32) \
				+ ((u64)desc->addr_lo & 0xffffffff)); \
} while (0)

#define discard_pkt() \
do { \
  forward_rx_ring_head(); \
  tw32_rx_mbox(MAILBOX_RCVRET_CON_IDX_0 + TG3_64BIT_REG_LOW, rx_ring_head); \
  if (opaque_key == RXD_OPAQUE_RING_STD) { \
    incr_modulo(*post_ptr,TG3_RX_RING_SIZE); \
    tw32_rx_mbox(MAILBOX_RCV_STD_PROD_IDX + TG3_64BIT_REG_LOW, *post_ptr); \
  } else if (opaque_key == RXD_OPAQUE_RING_JUMBO) { \
    incr_modulo(*post_ptr,TG3_RX_JUMBO_RING_SIZE); \
    tw32_rx_mbox(MAILBOX_RCV_JUMBO_PROD_IDX + TG3_64BIT_REG_LOW, *post_ptr); \
  } \
  mmiowb(); \
} while (0)

#define invoke_standard_rx(dev) tgg3_rx(netdev_priv(dev), 0)

#define rx_suspend_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		tgg3_disable_ints(netdev_priv(gamma_devices[i])); \
} while (0)

#define rx_resume_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		tgg3_enable_ints(netdev_priv(gamma_devices[i])); \
} while (0)
