#ifdef GAMMA_INSTALLING
#include "gamma_time.h"
#else
#include <linux/gamma/gamma_time.h>
#endif

#define DONT_MIND 0

#define IS_FIRST_CHUNK (unsigned long)0x40000000 /* bit 14 */
#define IS_SAME_NODE   (unsigned long)0x20000000 /* bit 13 */
#define PORT_MASK      (unsigned long)0x1fffffff /* bit 12 downto 0 */
/*  don't use the leftmost bit because it is sign!  < 0 means "err"  */

#define MAX_PORT_NUMBER 1024	/*
				 *  Max port number allowed by each GAMMA
				 *  process; ports are numbered from 0 to
				 *  MAX_PORT_NUMBER, so MAX_PORT_NUMBER+1 ports
				 *  per process.
				 *  Must not exceed 2^13-1
				 *  (useful bits from 12 downto 0, see above).
				 *  Port #MAX_PORT_NUMBER reserved to barrier.
				 *  Port #MAX_PORT_NUMBER-1 reserved to bcast.
				 */
#define MAX_SYNC_SEM_NUMBER	MAX_PORT_NUMBER
				/*
				 *  Max sync. semaphore number allowed by each
				 *  GAMMA process; they are numbered from 0 to
				 *  MAX_SYNC_SEM_NUMBER, so
				 *  MAX_SYNC_SEM_NUMBER+1 semaphores per
				 *  process.
				 *  Must not exceed 2^13-1
				 *  (useful bits from 12 downto 0, see above).
				 *  Sem #MAX_PORT_NUMBER reserved to barrier.
				 *  Sem #MAX_PORT_NUMBER-1 reserved to bcast.
				 */
#define MAX_PAR_PID 255		/*
				 *  Max par_pid number allowed.
				 *  Must not exceed the unsigned char range.
				 *  par_pid #0 is reserved.
				 */
#define MAX_NUM_NODES 128	/*
				 *  Max no. of machines in the cluster.
				 *  Must not exceed the unsigned short range.
				 *  Must be < BROADCAST (below).
				 *  The max no. of process instances is limited
				 *  by MAX_NUM_NODES * MAX_NUM_LOCAL_INST
				 */
#define BROADCAST 255		/*
				 *  A conventional node number
				 *  which denotes "all nodes".
				 *  Must not exceed the unsigned short range.
				 */
#define BCAST_ALL_NICS 31       /*
				 * special value meaning that message should
				 * be routed to all nics. (broadcast)
				 * Could not use BROADCAST because this value
				 * should be packaged in the isend_handle of
				 * non blocking routines, and so it should
				 * not exceed 2^5-1.
				 */


				/*  error codes  */
#define GAMMA_ERR_GENERIC		1
#define GAMMA_ERR_SENDFAILURE		2
#define GAMMA_ERR_TRYAGAIN		3
#define GAMMA_ERR_MUSTYIELD		4
#define GAMMA_ISEND_IS_DONE		5
#define GAMMA_RX_INCOMPLETE		6
#define GAMMA_MSG_INCOMPLETE		7
#define GAMMA_ERR_MAX			255  /*  No greater err codes!  */

/*
 *  These define the behaviour of GAMMA driver over user-provided buffers.
 *  GO_BACK means that each message directed to a buffer will be stored
 *  at its beginning.  If not otherwise programmed by a receiver handler,
 *  this implies that a subsequent message will overwrite the current one
 *  in the buffer.
 *  GO_AHEAD means that a subsequent message will be stored next and
 *  adjacent the current one in the buffer, unless otherwise specified by
 *  a receiver handler.
 */
/*  NONE OF THEM MAY BE 0  */
#define GO_BACK 1
#define GO_AHEAD 2

typedef struct {
	unsigned long addrl;
	unsigned short addrw;
	unsigned char addr[6];
} eth_addr;
